<?php

namespace Drupal\association;

use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\association\Entity\AssociationInterface;
use Drupal\association\Entity\AssociationPage;

/**
 * Default Entity Association Negotiator.
 *
 * The default association resolution searches the route for compatible entities
 * to find linked associations from, or associations directly from the route
 * parameters.
 */
class DefaultAssociationNegotiator implements AssociationNegotiatorInterface {

  const ROUTE_REGEX = '/^(entity|layout_builder.overrides)\.(association(?:_page)?)\.[a-z_]+$/';

  /**
   * An array of entity types IDs that can be linked to Entity Associations.
   *
   * @var string[]
   */
  protected $entityTypes;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Construct a new instance of the DefaultAssociationNegotiator class.
   *
   * @param \Drupal\association\EntityAdapterManagerInterface $entity_adapter_manager
   *   Manage entity types which are linkable to Entity Associations.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(EntityAdapterManagerInterface $entity_adapter_manager, EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypes = $entity_adapter_manager->getEntityTypes();
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function byRoute(RouteMatchInterface $route_match = NULL): ?AssociationInterface {
    if (!$route_match) {
      return NULL;
    }

    $params = $route_match->getParameters();

    if (preg_match(self::ROUTE_REGEX, $route_match->getRouteName())) {
      $entity = $params->get('association') ?: $params->get('association_page');

      if ($entity instanceof AssociationInterface) {
        return $entity;
      }
      elseif ($entity instanceof AssociationPage) {
        return $entity->getAssociation();
      }
    }

    // Check if entity types from route parameters have can have associations.
    foreach (array_intersect($params->keys(), $this->entityTypes) as $key) {
      $entity = $params->get($key);

      if ($entity instanceof ContentEntityInterface && $entity->hasField('associations')) {
        if ($assocLink = $entity->associations->entity) {
          return $assocLink->getAssociation();
        }
      }
    }

    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function byEntity(EntityInterface $entity): ?AssociationInterface {
    if ($entity instanceof AssociationLink) {
      $association = $entity;
    }
    elseif ($entity instanceof AssociationPage) {
      return $entity->getAssociation();
    }
    elseif (in_array($entity->getEntityTypeId(), $this->entityTypes) && $entity->hasField('associations')) {
      $association = $entity->associations->entity;
    }

    return $association ? $association->getAssociation() : NULL;
  }

}
