<?php

namespace Drupal\association\Entity\Form;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\path_alias\AliasRepositoryInterface;
use Drupal\pathauto\PathautoGeneratorInterface;

/**
 * Entity form for creating an association entity.
 */
class AssociationForm extends ContentEntityForm {

  /**
   * The path alias repository for getting the current path alias value.
   *
   * @var \Drupal\path_alias\AliasRepositoryInterface
   */
  protected $aliasRepo;

  /**
   * Pathauto generator to refresh path aliases when needed.
   *
   * @var \Drupal\pathauto\PathautoGeneratorInterface
   */
  protected $pathautoGenerator;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);

    $pathautoGenerator = $container->get('pathauto.generator', ContainerInterface::NULL_ON_INVALID_REFERENCE);
    if ($pathautoGenerator) {
      $instance->setPathautoGenerator($pathautoGenerator);
      $instance->setPathAliasRepository($container->get('path_alias.repository'));
    }

    return $instance;
  }

  /**
   * Sets the path alias repository if the service is available.
   *
   * @param \Drupal\path_alias\AliasRepositoryInterface $alias_repository
   *   The path alias repository for getting the current path alias value.
   */
  public function setPathAliasRepository(AliasRepositoryInterface $alias_repository) {
    $this->aliasRepo = $alias_repository;
  }

  /**
   * Sets the pathauto generator service.
   *
   * @param \Drupal\pathauto\PathautoGeneratorInterface $pathauto_generator
   *   Pathauto generator to refresh path aliases when needed.
   */
  public function setPathautoGenerator(PathautoGeneratorInterface $pathauto_generator) {
    $this->pathautoGenerator = $pathauto_generator;
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    $form['advanced'] = [
      '#type' => 'vertical_tabs',
      '#weight' => 50,
    ];

    $form['author'] = [
      '#type' => 'details',
      '#title' => $this->t('Authoring Information'),
      '#group' => 'advanced',
      '#optional' => TRUE,
      '#access' => $form['uid']['#access'] ?? NULL,
    ];

    $form['uid']['#group'] = 'author';
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $entity = $this->entity;
    $langcode = $this->getFormLangcode($form_state);
    $rebuildAliases = FALSE;

    // If this association has a path alias setup, then attempt to
    // update any sub-content path aliases.
    if (!$entity->isNew() && $this->pathautoGenerator && $entity->hasField('path')) {
      $alias = $entity->path->alias;
      $internalPath = '/' . $entity->toUrl()->getInternalPath();
      $currentAlias = $this->aliasRepo->lookupBySystemPath($internalPath, $langcode)['alias'] ?? '';
      $rebuildAliases = $alias != $currentAlias;
    }

    // Save the association entity, and check if aliases need updating.
    $status = parent::save($form, $form_state);
    $msgParams = [
      '%label' => $entity->label(),
      ':bundle' => $entity->getType()->label(),
    ];

    $msg = SAVED_NEW === $status
      ? $this->t('Created the %label :bundle.', $msgParams)
      : $this->t('Saved the %label :bundle.', $msgParams);

    $this->messenger()->addStatus($msg);
    $form_state->setRedirectUrl($entity->toUrl('manage'));

    // If the alias has changed, trigger a batch to rebuild all the URL aliases
    // for the association linked entity content.
    //
    // @todo should this get triggered when Association::postSave(), the issue
    // would be how to handle batching. Could implement as a worker queue?
    if ($rebuildAliases) {
      $operations = [];
      $operations[] = [
        static::class . '::updateLinkEntityAliases',
        [$entity->id()],
      ];

      batch_set([
        'title' => $this->t('Update content aliases'),
        'finished' => static::class . '::finishBatch',
        'operations' => $operations,
      ]);
    }

    return $status;
  }

  /**
   * Batch operation for updating association linked content path aliases.
   *
   * @param int $association_id
   *   The entity ID for the association that updated its path alias.
   * @param mixed $context
   *   The current batch context. The context maintains the results, sandbox
   *   and finished, which persist between the batch runs for this operation.
   */
  public static function updateLinkEntityAliases($association_id, &$context = []) {
    $storage = \Drupal::entityTypeManager()->getStorage('association_link');
    $query = $storage->getQuery()->condition('association', $association_id);

    // If operation is initializing, setup the total and progress tracking.
    if (!isset($context['sandbox']['progress'])) {
      $countQuery = clone $query;
      $context['results']['alias_updates'] = 0;
      $context['sandbox'] = [
        'current' => 0,
        'progress' => 0,
        'total' => $countQuery->count()->execute(),
      ];

      // No content to update, we're done here.
      if (!$context['sandbox']['total']) {
        return;
      }
    }

    $ids = $query
      ->condition('id', $context['sandbox']['current'], '>')
      ->sort('id', 'ASC')
      ->range(0, 50)
      ->execute();

    if ($ids) {
      $pathGenerator = \Drupal::service('pathauto.generator');

      // Update the linked pathauto url aliases.
      foreach ($storage->loadMultiple($ids) as $link) {
        $pathGenerator->updateEntityAlias($link->getTarget(), 'update');

        $context['sandbox']['current'] = $link->id();
        $context['sandbox']['progress']++;
        $context['results']['alias_updates']++;
      }

      // Update the overall batch process.
      if ($context['sandbox']['progress'] < $context['sandbox']['total']) {
        $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['total'];
      }
    }
  }

  /**
   * The association content URL alias update batch finished callback.
   *
   * @param bool $success
   *   Did the batch complete successfully.
   * @param array $results
   *   Results information from the processed operations.
   * @param array $operations
   *   The remaining unprocessed operations if they did not all complete.
   */
  public static function finishBatch($success, array $results, array $operations) {
    if ($success) {
      $msg = t('Updated @updated content aliases.', [
        '@updated' => $results['alias_updates'],
      ]);

      \Drupal::messenger()->addStatus($msg);
    }
  }

}
