<?php

namespace Drupal\association\Event;

use Drupal\Component\EventDispatcher\Event;
use Drupal\Core\Database\Query\AlterableInterface;
use Drupal\Core\Database\Query\ConditionInterface;
use Drupal\Core\Session\AccountInterface;

/**
 * Event interface for altering access queries criteria for associated entities.
 */
class AccessConditionAlterEvent extends Event implements AccessConditionAlterEventInterface {

  /**
   * The entity operation the query being built for.
   *
   * @var string
   */
  protected $operation;

  /**
   * Referece to the query condition which apply association access criteria.
   *
   * This is a reference to the original variable being worked on and can be
   * used to replace and alter the condition.
   *
   * This is important if it is currently NULL and access conditions need to
   * be added. In this case you replace it with a new condition group. Keep in
   * mind that in normal cases the expectation is that this is an "OR" group.
   *
   * @var \Drupal\Core\Database\Query\ConditionInterface|null
   */
  protected $accessCondition;

  /**
   * The query being altered.
   *
   * @var \Drupal\Core\Database\Query\AlterableInterface
   */
  protected $query;

  /**
   * The table aliases of the base table, associaton and association link.
   *
   * @var string[]
   */
  protected $tableAliases;

  /**
   * The account to use for access permission and access checks.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $account;

  /**
   * Create a new instance of the AccessConditionAlterEvent class.
   *
   * @param string $operation
   *   The entity operation being performed.
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query being altered.
   * @param \Drupal\Core\Database\Query\ConditionInterface|null $condition
   *   A reference to the query condition with the association access filters.
   * @param string[] $table_aliases
   *   The table aliases from the query for the entity base, association,
   *   and association link tables.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The account to use for access and permission checks.
   */
  public function __construct($operation, AlterableInterface $query, ?ConditionInterface &$condition, array $table_aliases, AccountInterface $account) {
    $this->operation = $operation;
    $this->query = $query;
    $this->accessCondition = &$condition;
    $this->tableAliases = $table_aliases;
    $this->account = $account;
  }

  /**
   * {@inheritdoc}
   */
  public function &getCondition(): ?ConditionInterface {
    return $this->accessCondition;
  }

  /**
   * {@inheritdoc}
   */
  public function getQuery(): AlterableInterface {
    return $this->query;
  }

  /**
   * {@inheritdoc}
   */
  public function getOperation() {
    return $this->operation;
  }

  /**
   * {@inheritdoc}
   */
  public function getTableAliases(): array {
    return $this->tableAliases;
  }

  /**
   * {@inheritdoc}
   */
  public function getAccount(): AccountInterface {
    return $this->account;
  }

}
