<?php

namespace Drupal\association\Event;

use Drupal\Core\Database\Query\AlterableInterface;
use Drupal\Core\Database\Query\ConditionInterface;
use Drupal\Core\Session\AccountInterface;

/**
 * Event interface for altering access queries criteria for associated entities.
 */
interface AccessConditionAlterEventInterface {

  const EVENT_NAME = 'association.access_condition_alter';

  /**
   * Gets the a reference to the access condition variable.
   *
   * This returns a reference to the condition variable that will be applied
   * to the query. This means that it's possible to completely replace this
   * or just add to it.
   *
   * Condition can be NULL, if there are currently no constraints, so adding
   * additional query conditions requires creating a new condition group first,
   * and replacing the condition.
   *
   * Keep in mind that the expectation is that this is an "OR" group.
   *
   * @return \Drupal\Core\Database\Query\ConditionInterface|null
   *   Get the current access condition with the association based access
   *   constraints applied. This can be NULL if there are currently no
   *   query constraints needed.
   */
  public function &getCondition(): ?ConditionInterface;

  /**
   * The query being altered.
   *
   * @return \Drupal\Core\Database\Query\AlterableInterface
   *   The query being altered.
   */
  public function getQuery(): AlterableInterface;

  /**
   * Gets the entity operation being performed.
   *
   * @return string
   *   The entity operation being performed.
   */
  public function getOperation();

  /**
   * Gets the table aliases of the entity base and association tables.
   *
   * @return string[]
   *   The table aliaes keyed by "base" (base entity table), "association"
   *   (association field data table), and "association_link" (association
   *   link base table).
   *
   * @see \Drupal\association\Adapte\AssociatedQueryAlterTrait::ensureTable()
   */
  public function getTableAliases(): array;

  /**
   * Get the account to use for permission and access checks.
   *
   * @return \Drupal\Core\Session\AccountInterface
   *   Gets the account to compare permission and access with.
   */
  public function getAccount(): AccountInterface;

}
