<?php

namespace Drupal\association\Event;

use Drupal\Component\EventDispatcher\Event;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\association\Entity\AssociationInterface;

/**
 * Event class for associated entity operations.
 */
class AssociatedEntityFormEvent extends Event implements AssociatedEntityFormEventInterface {

  /**
   * The event identifier for this current event.
   *
   * @var string
   */
  protected $event;

  /**
   * The entity assocation which the form is operating on.
   *
   * @var \Drupal\Core\Entity\AssociationInterface
   */
  protected $association;

  /**
   * The entity being operated on the the entity operation form.
   *
   * @var \Drupal\Core\Entity\EntityInterface
   */
  protected $entity;

  /**
   * Reference to the entity operation form elements.
   *
   * @var array
   */
  protected $form;

  /**
   * The entity operation form state.
   *
   * @var \Drupal\Core\Form\FormStateInterface
   */
  protected $formState;

  /**
   * Creates a new instance of the AssociationEntityFormEvent.
   *
   * @param string $op
   *   The entity form operation being performed.
   * @param array $form
   *   The reference to the form renderable form elements.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form rebuild info, state and values.
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity being edited by the form.
   * @param \Drupal\association\Entity\AssociationInterface $association
   *   The entity association to which this entity is a part of.
   */
  public function __construct($op, array &$form, FormStateInterface $form_state, ContentEntityInterface $entity, AssociationInterface $association) {
    $this->form = &$form;
    $this->formState = $form_state;
    $this->entity = $entity;
    $this->association = $association;

    $this->event = static::getEventNames()[$op] ?? NULL;
    if (!$this->event) {
      $error = sprintf('AssociatedEntityFormEvent does not support the %s entity form operation', $op);
      throw new \InvalidArgumentException($error);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function getEventNames(): array {
    return [
      'add' => self::INSERT_ENTITY,
      'edit' => self::UPDATE_ENTITY,
      'delete' => self::DELETE_ENTITY,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getEventName() {
    return $this->event;
  }

  /**
   * {@inheritdoc}
   */
  public function getEntity(): ContentEntityInterface {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getAssociation(): AssociationInterface {
    return $this->association;
  }

  /**
   * {@inheritdoc}
   */
  public function &getForm(): array {
    return $this->form;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormState(): FormStateInterface {
    return $this->formState;
  }

}
